#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#include "pairings.h"
#include "funcs.h"
#include "cyclo_poly.h"
#include "char_2/pairings_char_2.h"
#include "char_3/pairings_char_3.h"
#include "char_p/pairings_char_p.h"
#include "char_p/affine.h"
#include "char_p/proj.h"

#define __TIMING__ 1
#include "timer.h"
#if __TIMING__ == 1
    static void print_timing_msg(char* title, pairing_data_t* pdata, double timing);
    #define PRINT_TIMING_MSG(M, P, T) print_timing_msg(M, P, T);
#else
    #define PRINT_TIMING_MSG(M, P, T) /* voluntarily left empty */
#endif
//-----------------------------------------------------------------------------
pairing_data_t*
alloc_pairing_data(GEN curve, GEN charac,
                   GEN f1_defpol, GEN f2_defpol,
                   GEN g1_order, curve_family_t family)
{
    IN();

    GEN a1 = ell_get_a1(curve);
    GEN a2 = ell_get_a2(curve);
    GEN a3 = ell_get_a3(curve);
    if (!gequal0(a1) || !gequal0(a2) || !gequal0(a3)) {
        pariprintf("Curve must be of equation y^2 = x^3 + a.x + b.x\n");
        return NULL;
    }

    pairing_data_t* pdata = gpmalloc(sizeof(pairing_data_t));

    pdata->curve        = curve;
    pdata->charac       = charac;
    pdata->frob_trace   = NULL;
    pdata->f1_extdeg    = degpol(f1_defpol);
    pdata->f2_extdeg    = degpol(f2_defpol);
    pdata->irredpol     = gclone(f2_defpol);
    pdata->g1_order     = g1_order;
    pdata->embed_deg    = pdata->f2_extdeg / pdata->f1_extdeg;

    if (pdata->embed_deg > MAX_EMBED_DEGREE) {
        pariprintf(
            "Embedding degree should be less than %i\n", MAX_EMBED_DEGREE
        );
        gpfree(pdata);
        return NULL;
    }
    pdata->family       = family;
    pdata->coordsys     = AFFINE;
    pdata->final_exp    = gclone(
                              diviiexact(
                                  subis(powiu(charac, pdata->embed_deg), 1),
                                  g1_order
                              )
                          );
    pdata->miller_method  = STANDARD;
    pdata->hardpow_method = SCOTT_ET_AL;
    pdata->do_denom_elim = 0;
    pdata->precomp_done  = 0;
    pdata->fmaps_are_computed = 0;
    pdata->do_naive_exp  = 1;

    if (gequalgs(charac, 2)) {
        pdata->charac_code = CHARAC_2;
    } else if (gequalgs(charac, 3)) {
        pdata->charac_code = CHARAC_3;
    } else {
        pdata->charac_code = CHARAC_P;
    }

    pdata->char_is_1_mod_12  = (smodis(pdata->charac, 12) == 1);
    pdata->embed_deg_is_2i3j = 0;

    int k = pdata->embed_deg;
    while (IS_EVEN(k)) {
        k /= 2;
    }
    while (!(k % 3)) {
        k /= 3;
    }
    if (k == 1) {
        pdata->embed_deg_is_2i3j = 1;
    }

    pdata->fields_are_friendly =    pdata->embed_deg_is_2i3j
                                 && IS_EVEN(pdata->embed_deg);
    pdata->do_reduce = 1;

    k = pdata->embed_deg;

    GEN p   = pdata->charac;
    GEN phi = eval_ith_cyclotomic_poly(k, p);

    pdata->easy_exp_1 = NULL;
    pdata->easy_exp_2 = NULL;
    pdata->hard_exp   = NULL;

    pdata->edeg_sdiv = k;
    for (int i = 0; i < NTINY_PRIMES; i++) {
        if ((k % tiny_primes[i]) == 0) {
            pdata->edeg_sdiv = tiny_primes[i];
            break;
        }
    }
    int d = k / pdata->edeg_sdiv;
    pdata->easy_exp_1 = gclone(subis(powiu(p, d), 1));
    pdata->hard_exp   = gclone(diviiexact(phi, pdata->g1_order));
    pdata->easy_exp_2 = gclone(diviiexact(pdata->final_exp,
                                    mulii(pdata->easy_exp_1, pdata->hard_exp)));
    pdata->to_iso     = NULL;
    pdata->to_iso_inv = NULL;

    for (int i = 0; i <= MAX_NFROB_POWERS; i++) {
        pdata->frobenius_powers[i] = NULL;
    }
    pdata->u        = NULL;
    pdata->cof      = NULL;
    pdata->fe_coefs = NULL;

    pdata->do_naive_hard_exp = 0;
    update_pairing_data(pdata);

    pdata->r          = 0;
    pdata->twist_deg  = 0;
    pdata->card_aut_E = 0;

    pdata->frobcost = DFLT_FROB_COST;
    pdata->sqrcost  = DFLT_SQR_COST;

    GEN a4 = ell_get_a4(curve);
    GEN a6 = ell_get_a6(curve);
    if (!gequal0(a4) && !gequal0(a6)) {
        pdata->card_aut_E = 2;
    } else if (!gequal0(a4)) {
        pdata->card_aut_E = 4;
    } else if (!gequal0(a6)) {
        pdata->card_aut_E = 6;
    }

    OUT();
    return pdata;
}
//-----------------------------------------------------------------------------
void
free_pairing_data(pairing_data_t* pdata)
{
    gunclone(pdata->final_exp);
    gunclone(pdata->irredpol);
    if (pdata->easy_exp_1 != NULL) {
        gunclone(pdata->easy_exp_1);
    }
    if (pdata->easy_exp_2 != NULL) {
        gunclone(pdata->easy_exp_2);
    }
    if (pdata->hard_exp != NULL) {
        gunclone(pdata->hard_exp);
    }
    if (pdata->to_iso != NULL) {
        gunclone(pdata->to_iso);
    }
    if (pdata->to_iso_inv != NULL) {
        gunclone(pdata->to_iso_inv);
    }
    for (int i = 0; i < MAX_NFROB_POWERS; i++) {
        if (pdata->frobenius_powers[i] != NULL) {
            gunclone(pdata->frobenius_powers[i]);
        }
    }
    if (pdata->u != NULL) {
        gunclone(pdata->u);
    }
    if (pdata->cof != NULL) {
        gunclone(pdata->cof);
    }
    if (pdata->fe_coefs != NULL) {
        gunclone(pdata->fe_coefs);
    }
    gpfree(pdata);
}
//-----------------------------------------------------------------------------
void
update_pairing_data(pairing_data_t* pdata)
{
    switch (pdata->charac_code) {
        case CHARAC_2:
            return update_pairing_data_char_2(pdata);

        case CHARAC_3:
            return update_pairing_data_char_3(pdata);

        case CHARAC_P:
            return update_pairing_data_char_p(pdata);

        default:
            break;
    }
}
//-----------------------------------------------------------------------------
void set_twist_degree(pairing_data_t* pdata, unsigned int d)
{
    pdata->twist_deg = 0;
    switch (pdata->card_aut_E) {
        case 2:
            if (d == 2) {
                if (umodiu(pdata->charac, 2) == 1) {
                    pdata->twist_deg = 2;
                }
            }
            break;
        case 4:
            if (d == 4) {
                if (umodiu(pdata->charac, 4) == 1) {
                    pdata->twist_deg = 4;
                }
            }
            break;
        case 6:
            if (d == 3) {
                if (umodiu(pdata->charac, 3) == 1) {
                    pdata->twist_deg = 3;
                }
            }
            if (d == 6) {
                if (umodiu(pdata->charac, 6) == 1) {
                    pdata->twist_deg = 6;
                }
            }
            break;
        default:
            break;
    }
    update_pairing_data(pdata);
    if (pdata->twist_deg == 0) {
        pariprintf("Impossible twist degree\n");
    }
    return;
}
//-----------------------------------------------------------------------------
GEN
tate(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Tate pairing t(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // If the characteristic of the involved field is p > 3 then:
    //
    //     - P must be a point on E(F_p) of order r whose coordinates are
    //       given by t_INTMODs.
    //     - Q must be a point on E(F_{p^k}) of order r whose coordinates
    //       are given by t_FFELTs of format t_FF_FpXQ or t_FF_Flxq.
    //
    // WARNING: Characteristic 2 and 3 are currently not implemented!
    //
    IN();
    INIT_TIMER;
    START_TIMER;
    
    pari_sp ltop = avma;
    if (ell_is_inf(P) || ell_is_inf(Q)) {
        return gen_1;
    }
    GEN p = pdata->charac;
    GEN res = NULL;

    if (gequalgs(p, 2)) {
        // To be implemented
    } else if (gequalgs(p, 3)) {
        // To be implemented
    } else {
        res = tate_pairing_char_p(pdata, P, Q);
    }
    STOP_TIMER;
    PRINT_TIMING_MSG("tate", pdata, GET_TIMING);
    OUT();
    return gerepileupto(ltop, res);
}
//-----------------------------------------------------------------------------
GEN
weil(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Weil pairing t(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    //
    // P and Q are two points of order r. Two cases are currently implemented:
    //
    //     a) P is a point on E(F_p) of order r whose coordinates are given by
    //        t_INTMODs and Q is a point on E(F_{p^k}) of order r whose
    //        coordinates are given by t_FFELTs
    // or
    //
    //     b) P is a point on E(F_{p^k}) of order r whose coordinates are given
    //        by t_FFELTs and Q is a point on E(F_p) of order r whose
    //        coordinates are given by t_INTMODs
    //
    // This function correctly cleans the stack upon completion.
    //
    IN();
    INIT_TIMER;
    START_TIMER;
    
    pari_sp ltop = avma;

    if (ell_is_inf(P) || ell_is_inf(Q)) {
        return gen_1;
    }

    GEN p = pdata->charac;
    GEN res = NULL;

    if (gequalgs(p, 2)) {
        // To be implemented
    } else if (gequalgs(p, 3)) {
        // To be implemented
    } else {
        res = weil_pairing_char_p(pdata, P, Q);
    }
    STOP_TIMER;
    PRINT_TIMING_MSG("weil", pdata, GET_TIMING);
    OUT();
    
    return gerepileupto(ltop, res);
}
//-----------------------------------------------------------------------------
GEN
ate(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Ate pairing a(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    //     t = pdata->frob_trace
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // If the characteristic of the involved field is p > 3 then:
    //
    //     - P must be a point on E(F_p) of order r (a factor of #E(F_p))
    //       whose coordinates are given by t_INTMODs.
    //     - Q must be a point on E(F_{p^k}) of order r whose coordinates
    //       are given by t_FFELTs of format t_FF_FpXQ or t_FF_Flxq.
    //
    // WARNING: Characteristic 2 and 3 are currently not implemented!
    //
    IN();
    INIT_TIMER;
    START_TIMER;
    
    pari_sp ltop = avma;

    if (ell_is_inf(P) || ell_is_inf(Q)) {
        return gen_1;
    }

    GEN p = pdata->charac;
    GEN res = NULL;

    if (gequalgs(p, 2)) {
        // To be implemented
    } else if (gequalgs(p, 3)) {
        // To be implemented
    } else {
        res = ate_pairing_char_p(pdata, P, Q);
    }
    STOP_TIMER;
    PRINT_TIMING_MSG("ate", pdata, GET_TIMING);
    OUT();
    return gerepileupto(ltop, res);
}
//-----------------------------------------------------------------------------
GEN
twisted(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Twisted Ate pairing twist(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // If the characteristic of the involved field is p > 3 then:
    //
    //     - P must be a point on E(F_p) of order r whose coordinates are
    //       given by t_INTMODs.
    //     - Q must be a point on E(F_{p^k}) of order r whose coordinates
    //       are given by t_FFELTs of format t_FF_FpXQ or t_FF_Flxq.
    //
    // WARNING: Characteristic 2 and 3 are currently not implemented!
    //
    IN();
    INIT_TIMER;
    START_TIMER;
    
    pari_sp ltop = avma;
    if (ell_is_inf(P) || ell_is_inf(Q)) {
        return gen_1;
    }
    GEN p = pdata->charac;
    GEN res = NULL;

    if (gequalgs(p, 2)) {
        // To be implemented
    } else if (gequalgs(p, 3)) {
        // To be implemented
    } else {
        res = twisted_pairing_char_p(pdata, P, Q);
    }
    STOP_TIMER;
    PRINT_TIMING_MSG("twisted", pdata, GET_TIMING);
    OUT();
    return gerepileupto(ltop, res);
}
//-----------------------------------------------------------------------------
GEN
opti_ate(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Optimal Ate pairing a(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    //     t = pdata->frob_trace
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // If the characteristic of the involved field is p > 3 then:
    //
    //     - P must be a point on E(F_p) of order r (a factor of #E(F_p))
    //       whose coordinates are given by t_INTMODs.
    //     - Q must be a point on E(F_{p^k}) of order r whose coordinates
    //       are given by t_FFELTs of format t_FF_FpXQ or t_FF_Flxq.
    //
    // WARNING: Characteristic 2 and 3 are currently not implemented!
    //
    IN();
    INIT_TIMER;
    START_TIMER;
    
    pari_sp ltop = avma;
    if (ell_is_inf(P) || ell_is_inf(Q)) {
        return gen_1;
    }

    GEN p = pdata->charac;
    GEN res = NULL;

    if (gequalgs(p, 2)) {
        // To be implemented
    } else if (gequalgs(p, 3)) {
        // To be implemented
    } else {
        res = opti_ate_pairing_char_p(pdata, P, Q);
    }
    STOP_TIMER;
    PRINT_TIMING_MSG("opti ate", pdata, GET_TIMING);
    OUT();
    return gerepileupto(ltop, res);
}
//-----------------------------------------------------------------------------
GEN
opti_twisted(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Optimal twisted ate pairing a(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    //     t = pdata->frob_trace
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // If the characteristic of the involved field is p > 3 then:
    //
    //     - P must be a point on E(F_p) of order r (a factor of #E(F_p))
    //       whose coordinates are given by t_INTMODs.
    //     - Q must be a point on E(F_{p^k}) of order r whose coordinates
    //       are given by t_FFELTs of format t_FF_FpXQ or t_FF_Flxq.
    //
    // WARNING: Characteristic 2 and 3 are currently not implemented!
    //
    IN();
    INIT_TIMER;
    START_TIMER;
    
    pari_sp ltop = avma;
    if (ell_is_inf(P) || ell_is_inf(Q)) {
        return gen_1;
    }

    GEN p = pdata->charac;
    GEN res = NULL;

    if (gequalgs(p, 2)) {
        // To be implemented
    } else if (gequalgs(p, 3)) {
        // To be implemented
    } else {
        res = opti_twisted_pairing_char_p(pdata, P, Q);
    }
    STOP_TIMER;
    PRINT_TIMING_MSG("opti twisted", pdata, GET_TIMING);
    OUT();
        
    return gerepileupto(ltop, res);
}
//-----------------------------------------------------------------------------
GEN
miller(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    IN();
    INIT_TIMER;
    START_TIMER;
    pari_sp ltop = avma;
    if (ell_is_inf(P) || ell_is_inf(Q)) {
        return gen_1;
    }
    GEN p = pdata->charac;
    GEN res = NULL;

    if (gequalgs(p, 2)) {
        // To be implemented
    } else if (gequalgs(p, 3)) {
        // To be implemented
    } else {
        res = miller_char_p(pdata, r, P, Q);
    }
    STOP_TIMER;
    PRINT_TIMING_MSG("miller", pdata, GET_TIMING);
    OUT();
    
    return gerepileupto(ltop, res);
}
//-----------------------------------------------------------------------------
void
compute_frobenius_powers(pairing_data_t* const pdata)
{
    IN();
    INIT_TIMER;
    START_TIMER;
    
    GEN p = pdata->charac;
    if (gequalgs(p, 2)) {
        // To be implemented
    } else if (gequalgs(p, 3)) {
        // To be implemented
    } else {
        compute_frobenius_powers_char_p(pdata);
    }
    STOP_TIMER;
    PRINT_TIMING("comp_frob_powers");
    OUT();
}
//-----------------------------------------------------------------------------
GEN
frobenius_power(pairing_data_t* const pdata, GEN X, unsigned int i)
{
    IN();
    pari_sp ltop = avma;
    GEN p = pdata->charac;
    GEN res = NULL;
    if (gequalgs(p, 2)) {
        // To be implemented
    } else if (gequalgs(p, 3)) {
        // To be implemented
    } else {
        res = frobenius_power_char_p(pdata, X, i);
    }
    OUT();
    return gerepileupto(ltop, res);
}
//-----------------------------------------------------------------------------
void
print_frobenius_matrix(pairing_data_t* const pdata, unsigned int i)
{
    IN();
    GEN p = pdata->charac;
    if (gequalgs(p, 2)) {
        // To be implemented
    } else if (gequalgs(p, 3)) {
        // To be implemented
    } else {
        print_frobenius_matrix_char_p(pdata, i);
    }
    OUT();
    return;
}
//-----------------------------------------------------------------------------
static void
print_timing_msg(char* title, pairing_data_t* pdata, double timing) {
    char fulltitle[64];
    fulltitle[0] = '\0';
    (void)strncat(fulltitle, title, 12);
    if (pdata->do_reduce) {
        (void)strncat(fulltitle, "-red", 5);
    } else {
        (void)strncat(fulltitle, "-unred", 7);
    }
    if (pdata->do_naive_exp) {
        (void)strncat(fulltitle, "-naive", 7);
    } else {
        (void)strncat(fulltitle, "-opti", 6);
    }
    switch (pdata->miller_method) {
        case STANDARD:
            (void)strncat(fulltitle, "-std", 5);
            break;
        case NAF:
            (void)strncat(fulltitle, "-naf", 5);
            break;
        case BOXALL_ET_AL:
            (void)strncat(fulltitle, "-box", 5);
            break;
        case BOXALL_ET_AL_NAF:
            (void)strncat(fulltitle, "-box-naf", 8);
            break;
        default:
            (void)strncat(fulltitle, "-?", 3);
            break;
    }
    switch (pdata->coordsys) {
         case JACOBIAN:
            (void)strncat(fulltitle, "-jac", 5);
            break;
         case PROJECTIVE:
            (void)strncat(fulltitle, "-proj", 6);
            break;
         case AFFINE:
            (void)strncat(fulltitle, "-aff", 5);
            break;
         default:
            (void)strncat(fulltitle, "-?", 3);
            break;
    }
    printf("%s : "TIMING_FORMAT"\n", fulltitle, timing);
}
//-----------------------------------------------------------------------------
