#if !defined(_APIP_STATS_H_)
#define _APIP_STATS_H_

#include "apip_config.h"

#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#include "funcs.h"

#ifdef __cplusplus
extern "C" {
#endif

//
// Counters to keep track of the number of operations performed.
// 
// _WARNING_: Static variables do NOT work in this particular case as we would
//            end up with MULTIPLES versions of them, each file having its own
//            copies. This is NOT what we want.
//
extern ulong g1_g1_add;
extern ulong g1_g2_add;
extern ulong g2_g2_add;
extern ulong g1_g1_mul;
extern ulong g1_sqr;
extern ulong g1_g2_mul;
extern ulong g2_g2_mul;
extern ulong g2_sqr;
extern ulong g1_inv;
extern ulong g2_inv;
extern ulong g2_frob_pows;
extern ulong edegree;

extern float sqr_to_mul_ratio_g1;
extern float sqr_to_mul_ratio_g2;
extern float inv_to_mul_ratio_g1;
extern float inv_to_mul_ratio_g2;

//
// Default verbosity level to use.
//
extern ulong stats_verbosity;

//
// Should we compute and print statistics?
//
#define SHOULD_PRINT_STATS 0
                           //APIP_PRINT_STATS
//
// Verbosity level -- print every operation counts.
//
#define STATS_DETAILLED     4
//
// Verbosity level -- merge operation counts when possible (e.g. consider a 
// G1-G2 multiplication as several G1-G1 multiplication).
//
#define STATS_SIMPLIFIED    3
//
// Verbosity level -- merge operation counts and ignore additions.
//
#define STATS_SHORT         2
//
// Verbosity level -- aggresively merge operation counts and ignore additions.
//
#define STATS_SHORTEST      1
//
// Verbosity option -- subtracts the cost of all Frobenius powers.
//
#define SUBTRACT_FROB_COST  0
//
// Default verbosity level to use.
//
#define DFLT_STATS_VERBOSITY     STATS_SHORTEST
//
// Should we ignore multiplications by small constants (like 2, 4 or 8) and
// replace them by additions in the statistics ?
//
#define IGNORE_MULT_BY_SMALL_CONST 1
//
// Default values for operations equaivalences
//
#define DFLT_SQR_TO_MUL_RATIO_G1   0.9
#define DFLT_SQR_TO_MUL_RATIO_G2   0.9
#define DFLT_INV_TO_MUL_RATIO_G1  12.0
#define DFLT_INV_TO_MUL_RATIO_G2   8.0

#if SHOULD_PRINT_STATS != 0
    #define PRINT_STATS(MSG) do {                       \
            pariprintf("    --------------------\n");   \
            pariprintf("    %s\n", MSG);                \
            pariprintf("    --------------------\n");   \
            print_stats();                              \
        } while (0)
    #define PRINT_RATIOS()          print_ratios();
    #define RESET_STATS()           reset_stats()
    #define SET_STATS_EDEGREE(X)    edegree = (X);
#else
    #define PRINT_STATS(MSG)        /* Intentionally left empty */
    #define PRINT_RATIOS()          /* Intentionally left empty */
    #define RESET_STATS()           /* Intentionally left empty */
    #define SET_STATS_EDEGREE(X)    /* Intentionally left empty */
#endif

//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE void
reset_stats() {
    g1_g1_add = 0;
    g1_g2_add = 0;
    g2_g2_add = 0;
    
    g1_g1_mul = 0;
    g1_g2_mul = 0;
    g2_g2_mul = 0;
    
    g1_sqr = 0;
    g2_sqr = 0;
    
    g1_inv = 0;
    g2_inv = 0;
    
    g2_frob_pows = 0;
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE void
print_stats() {
    #if SUBTRACT_FROB_COST
        ulong g1_g1_mul_bck = g1_g1_mul;
        ulong g1_g1_add_bck = g1_g1_add;
        g1_g1_mul -= g2_frob_pows * edegree * edegree;
        g1_g1_add -= g2_frob_pows * edegree * (edegree - 1);
    #endif
    
    switch (stats_verbosity) {
    
      case STATS_DETAILLED:
          pariprintf("    G1 G1 adds : %6lu\n", g1_g1_add);
          pariprintf("    G1 G2 adds : %6lu\n", g1_g2_add);
          pariprintf("    G2 G2 adds : %6lu\n\n", g2_g2_add);
          
          pariprintf("    G1 G1 muls : %6lu\n", g1_g1_mul);
          pariprintf("    G1 G2 muls : %6lu\n", g1_g2_mul);
          pariprintf("    G2 G2 muls : %6lu\n\n", g2_g2_mul);
          
          pariprintf("       G1 sqrs : %6lu\n", g1_sqr);
          pariprintf("       G2 sqrs : %6lu\n\n", g2_sqr);
          
          pariprintf("       G1 invs : %6lu\n", g1_inv);
          pariprintf("       G2 invs : %6lu\n\n", g2_inv);
          break;
      
      case STATS_SIMPLIFIED:
          pariprintf(
              "    G1 G1 adds : %6lu (est.)\n\n",
              g1_g1_add + g1_g2_add + edegree * g2_g2_add
          );
          pariprintf(
              "    G1 G1 muls : %6lu (est.)\n", g1_g1_mul + edegree * g1_g2_mul
          );
          
          pariprintf("    G2 G2 muls : %6lu\n\n", g2_g2_mul);
          
          pariprintf("       G1 sqrs : %6lu\n", g1_sqr);
          pariprintf("       G2 sqrs : %6lu\n\n", g2_sqr);
          
          pariprintf("       G1 invs : %6lu\n", g1_inv);
          pariprintf("       G2 invs : %6lu\n\n", g2_inv);
          break;
      
      case STATS_SHORT:
          pariprintf(
              "    G1 G1 muls : %6lu (est.)\n",
              g1_g1_mul + edegree * g1_g2_mul
              + (ulong)(sqr_to_mul_ratio_g1 * g1_sqr)
          );
          pariprintf(
              "    G2 G2 muls : %6lu (est.)\n",
              g2_g2_mul + (ulong)(sqr_to_mul_ratio_g2 * g2_sqr)
          );
          pariprintf("       G1 invs : %6lu\n", g1_inv);
          pariprintf("       G2 invs : %6lu\n\n", g2_inv);
          break;
      
      case STATS_SHORTEST:
      default:
          pariprintf(
              "    G1 G1 muls : %6lu (est.)\n",
              g1_g1_mul + edegree * g1_g2_mul
              + (ulong)(sqr_to_mul_ratio_g1 * g1_sqr)
              + (ulong)(inv_to_mul_ratio_g1 * g1_inv)
          );
          pariprintf(
              "    G2 G2 muls : %6lu (est.)\n\n",
              g2_g2_mul + (ulong)(sqr_to_mul_ratio_g2 * g2_sqr)
              + (ulong)(inv_to_mul_ratio_g2 * g2_inv)
          );
    }
    #if SUBTRACT_FROB_COST
        g1_g1_mul = g1_g1_mul_bck;
        g1_g1_add = g1_g1_add_bck;
        pariprintf("     Frob pows : %6lu\n\n", g2_frob_pows);
    #else
        if (g2_frob_pows != 0) {
            pariprintf("   (includes %lu Frob powers)\n\n", g2_frob_pows);
            pariflush();
        }
    #endif
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE void
print_ratios() {
    if (stats_verbosity > STATS_SHORT) {
        //
        // All operations are printed if stats_verbosity > STATS_SHORT,
        // so we don't make estimations using the S / M and I / M ratios.
        //
        return;
    }
    pariprintf("    --------------------\n");
    pariprintf("    Used ratios\n");
    pariprintf("    --------------------\n");
    pariprintf("    G1 : S / M = %5.2f\n",   sqr_to_mul_ratio_g1);
    pariprintf("         I / M = %5.2f\n",   inv_to_mul_ratio_g1);
    pariprintf("    G2 : S / M = %5.2f\n",   sqr_to_mul_ratio_g2);
    pariprintf("         I / M = %5.2f\n\n", inv_to_mul_ratio_g2);
    pariflush();
}
//----------------------------------------------------------------------------

#ifdef __cplusplus
}
#endif
#endif
